from hotspots import calculation


__author__ = "Chris Radoux, Peter Curran, Mihaela Smilova"
__copyright__ = None
__credits__ = None
__license__ = None
__version__ = "1.0.0"
__maintainer__ = "Peter Curran"
__email__ = "pcurran@ccdc.cam.ac.uk"
__status__ = "Development"
name = "hotspots"
